"""
PDB molecular container
=======================

Molecular container for storing all contents of PDB files.
"""
import logging
import os
from typing import Dict, List, Optional, Tuple
from propka.parameters import Parameters

import propka.version
from propka.output import write_pka, print_header, print_result
from propka.conformation_container import ConformationContainer
from propka.lib import make_grid, Options


_LOGGER = logging.getLogger(__name__)


class MolecularContainer:
    """Container for storing molecular contents of PDB files.

    TODO - this class name does not conform to PEP8 but has external use.
    We should deprecate and change eventually.


    .. versionchanged:: 3.4.0
       Removed :meth:`write_propka` and
       :meth:`additional_setup_when_reading_input_file` as reading and writing
       PROPKA input files is no longer supported.
    """

    conformation_names: List[str]
    conformations: Dict[str, ConformationContainer]
    name: Optional[str]
    version: propka.version.Version

    def __init__(self, parameters: Parameters, options: Options) -> None:
        """Initialize molecular container.

        Args:
            parameters:  Parameters() object
            options:  options object
        """
        # printing out header before parsing input
        print_header()
        self.conformation_names = []
        self.conformations = {}
        self.options = options
        self.name = None
        try:
            version_class = getattr(propka.version, parameters.version)
            self.version = version_class(parameters)
        except AttributeError as err:
            print(err)
            errstr = 'Error: Version {0:s} does not exist'.format(
                parameters.version)
            raise Exception(errstr)

    def top_up_conformations(self) -> None:
        """Makes sure that all atoms are present in all conformations."""
        ref_atoms = {
            atom.residue_label: atom
            for name in reversed(self.conformation_names)
            for atom in self.conformations[name].atoms
        }
        for conf in self.conformations.values():
            conf.top_up_from_atoms(ref_atoms.values())

    def find_covalently_coupled_groups(self) -> None:
        """Find covalently coupled groups."""
        for name in self.conformation_names:
            self.conformations[name].find_covalently_coupled_groups()

    def find_non_covalently_coupled_groups(self) -> None:
        """Find non-covalently coupled groups."""
        verbose = self.options.display_coupled_residues
        for name in self.conformation_names:
            self.conformations[name].find_non_covalently_coupled_groups(
                verbose=verbose)

    def extract_groups(self) -> None:
        """Identify the groups needed for pKa calculation."""
        for name in self.conformation_names:
            self.conformations[name].extract_groups()

    def calculate_pka(self) -> None:
        """Calculate pKa values."""
        # calculate for each conformation
        for name in self.conformation_names:
            self.conformations[name].calculate_pka(
                self.version, self.options)
        # find non-covalently coupled groups
        self.find_non_covalently_coupled_groups()
        # find the average of the conformations
        self.average_of_conformations()
        # print out the conformation-average results
        print_result(self, 'AVR', self.version.parameters)

    def average_of_conformations(self) -> None:
        """Generate an average of conformations."""
        parameters = self.conformations[self.conformation_names[0]].parameters
        # make a new configuration to hold the average values
        avr_conformation = ConformationContainer(
            name='average', parameters=parameters, molecular_container=self)
        container = self.conformations[self.conformation_names[0]]
        for group in container.get_groups_for_calculations():
            # new group to hold average values
            avr_group = group.clone()
            # sum up all groups ...
            for name in self.conformation_names:
                group_to_add = self.conformations[name].find_group(group)
                if group_to_add:
                    avr_group += group_to_add
                else:
                    str_ = (
                        'Group {0:s} could not be found in '
                        'conformation {1:s}.'.format(
                            group.atom.residue_label, name))
                    _LOGGER.warning(str_)
            # ... and store the average value
            avr_group = avr_group / len(self.conformation_names)
            avr_conformation.groups.append(avr_group)
        # store information on coupling in the average container
        if len(list(filter(lambda c: c.non_covalently_coupled_groups,
                           self.conformations.values()))):
            avr_conformation.non_covalently_coupled_groups = True
        # store chain info
        avr_conformation.chains = self.conformations[
            self.conformation_names[0]].chains
        self.conformations['AVR'] = avr_conformation

    def write_pka(self, filename=None, reference="neutral",
                  direction="folding", options=None) -> None:
        """Write pKa information to a file.

        Args:
            filename:  file to write to
            reference:  reference state
            direction:  folding vs. unfolding
            options:  options object
        """
        if filename is None:
            filename = os.path.join('{0:s}.pka'.format(self.name))
        # if the display_coupled_residues option is true, write the results out
        # to an alternative pka file
        if self.options.display_coupled_residues:
            filename = os.path.join('{0:s}_alt_state.pka'.format(self.name))
        if (hasattr(self.version.parameters, 'output_file_tag')
                and len(self.version.parameters.output_file_tag) > 0):
            filename = os.path.join(
                '{0:s}_{1:s}.pka'.format(
                    self.name, self.version.parameters.output_file_tag))
        write_pka(
            self, self.version.parameters, filename=filename,
            conformation='AVR', reference=reference)

    def get_folding_profile(self, conformation='AVR', reference="neutral",
                            grid: Tuple[float, float, float] = (0., 14., 0.1)):
        """Get a folding profile.

        Args:
            conformation:  conformation to select
            reference:  reference state
            direction:  folding direction (folding)
            grid:  the grid of pH values [min, max, step_size]
            options:  options object
        Returns:
            TODO - figure out what these are
            1. profile
            2. opt
            3. range_80pct
            4. stability_range
        """
        # calculate stability profile
        profile: List[Tuple[float, float]] = []
        for ph in make_grid(*grid):
            conf = self.conformations[conformation]
            ddg = conf.calculate_folding_energy(ph=ph, reference=reference)
            profile.append((ph, ddg))
        # find optimum
        opt: Tuple[Optional[float], float] = (None, 1e6)
        for point in profile:
            opt = min(opt, point, key=lambda v: v[1])
        # find values within 80 % of optimum
        range_80pct: Tuple[Optional[float], Optional[float]] = (None, None)
        values_within_80pct = [p[0] for p in profile if p[1] < 0.8*opt[1]]
        if len(values_within_80pct) > 0:
            range_80pct = (min(values_within_80pct), max(values_within_80pct))
        # find stability range
        stability_range: Tuple[Optional[float], Optional[float]] = (None, None)
        stable_values = [p[0] for p in profile if p[1] < 0.0]
        if len(stable_values) > 0:
            stability_range = (min(stable_values), max(stable_values))
        return profile, opt, range_80pct, stability_range

    def get_charge_profile(self, conformation: str = 'AVR', grid=(0., 14., .1)):
        """Get charge profile for conformation as function of pH.

        Args:
            conformation:  conformation to test
            grid:  grid of pH values [min, max, step]
        Returns:
            list of charge state values
        """
        charge_profile: List[List[float]] = []
        for ph in make_grid(*grid):
            conf = self.conformations[conformation]
            q_unfolded, q_folded = conf.calculate_charge(
                self.version.parameters, ph=ph)
            charge_profile.append([ph, q_unfolded, q_folded])
        return charge_profile

    def get_pi(self, conformation: str = 'AVR', grid=(0., 14.), *,
               precision: float = 1e-4) -> Tuple[float, float]:
        """Get the isoelectric points for folded and unfolded states.

        Args:
            conformation:  conformation to test
            grid:  pH window [min, max]
            precision:  Compute pI up to this precision
        Returns:
            1. Folded state PI
            2. Unfolded state PI
        """
        conf = self.conformations[conformation]

        WHICH_UNFOLDED = 0
        WHICH_FOLDED = 1

        def pi(which, pH, min_, max_):
            charge = conf.calculate_charge(
                self.version.parameters, ph=pH)[which]
            if max_ - min_ > precision:
                if charge > 0.0:
                    min_ = pH
                else:
                    max_ = pH
                next_pH = (min_ + max_) / 2
                return pi(which, next_pH, min_, max_)
            return pH

        start = (grid[0] + grid[1]) / 2, grid[0], grid[1]

        return (
            pi(WHICH_FOLDED, *start),
            pi(WHICH_UNFOLDED, *start),
        )



"""
Calculations
============

Mathematical helper functions.
"""

import math
from typing import Iterable, Optional, Tuple, TypeVar
from .vector_algebra import _XYZ

_BoundXYZ_1 = TypeVar("_BoundXYZ_1", bound=_XYZ)
_BoundXYZ_2 = TypeVar("_BoundXYZ_2", bound=_XYZ)

#: Maximum distance used to bound calculations of smallest distance
MAX_DISTANCE = 1e6


def squared_distance(atom1: _XYZ, atom2: _XYZ) -> float:
    """Calculate the squared distance between two atoms.

    Args:
        atom1:  first atom for distance calculation
        atom2:  second atom for distance calculation
    Returns:
        distance squared
    """
    dx = atom2.x - atom1.x
    dy = atom2.y - atom1.y
    dz = atom2.z - atom1.z
    res = dx*dx+dy*dy+dz*dz
    return res


def distance(atom1: _XYZ, atom2: _XYZ) -> float:
    """Calculate the distance between two atoms.

    Args:
        atom1:  first atom for distance calculation
        atom2:  second atom for distance calculation
    Returns:
        distance
    """
    return math.sqrt(squared_distance(atom1, atom2))


def get_smallest_distance(
    atoms1: Iterable[_BoundXYZ_1],
    atoms2: Iterable[_BoundXYZ_2],
) -> Tuple[Optional[_BoundXYZ_1], float, Optional[_BoundXYZ_2]]:
    """Calculate the smallest distance between two groups of atoms.

    Args:
        atoms1:  atom group 1
        atoms2:  atom group 2
    Returns:
        smallest distance between groups
    """
    res_dist = MAX_DISTANCE
    res_atom1 = None
    res_atom2 = None
    for atom1 in atoms1:
        for atom2 in atoms2:
            dist = squared_distance(atom1, atom2)
            if dist < res_dist:
                res_dist = dist
                res_atom1 = atom1
                res_atom2 = atom2
    return (res_atom1, math.sqrt(res_dist), res_atom2)



"""
Output
======

Output routines.


.. versionchanged::3.4.0
   Removed :func:`write_proka` as writing PROPKA input files is no longer
   supported.
"""
import logging
from datetime import date
from decimal import Decimal
from os import PathLike
from pathlib import Path
from typing import IO, AnyStr, List, Optional, Tuple, Union, TYPE_CHECKING
import warnings

from .parameters import Parameters
from . import __version__

if TYPE_CHECKING:
    from .atom import Atom
    from .conformation_container import ConformationContainer
    from .molecular_container import MolecularContainer

# https://docs.python.org/3/glossary.html#term-path-like-object
_PathLikeTypes = (PathLike, str)
_PathArg = Union[PathLike, str]
_IOSource = Union[IO[AnyStr], PathLike, str]
_TextIOSource = _IOSource[str]

_LOGGER = logging.getLogger(__name__)


def open_file_for_writing(input_file: _TextIOSource) -> IO[str]:
    """Open file or file-like stream for writing.

    TODO - convert this to a context manager.

    Args:
        input_file: path to file or file-like object. If file-like object,
        then will attempt to get file mode.
    """
    if isinstance(input_file, _PathLikeTypes):
        return open(input_file, 'wt')

    if not input_file.writable():
        raise IOError("File/stream not open for writing")

    return input_file


def print_header():
    """Print header section of output."""
    str_ = "{0:s}\n".format(get_propka_header())
    str_ += "{0:s}\n".format(get_references_header())
    str_ += "{0:s}\n".format(get_warning_header())
    _LOGGER.info("\n%s", str_)


def write_pdb_for_conformation(conformation: "ConformationContainer",
                               filename: _PathArg):
    """Write PDB conformation to a file.

    Args:
        conformation:  conformation container
        filename:  filename for output
    """
    warnings.warn("unused and untested by propka")
    write_pdb_for_atoms(conformation.atoms, filename)


def write_pka(protein: "MolecularContainer",
              parameters: Parameters,
              filename: Optional[_PathArg] = None,
              conformation: str = '1A',
              reference: str = "neutral",
              *,
              verbose: bool = True):
    """Write the pKa-file based on the given protein.

    Args:
        protein:  protein object
        filename:  output file name
        conformation: specific conformation
        reference:  reference state
        verbose:  Boolean flag for verbosity
    """
    if filename is None:
        filename = "{0:s}.pka".format(protein.name)
    if verbose:
        _LOGGER.info("Writing {0:s}".format(filename))
    # writing propka header
    str_ = "{0:s}\n".format(get_propka_header())
    str_ += "{0:s}\n".format(get_references_header())
    str_ += "{0:s}\n".format(get_warning_header())
    # writing pKa determinant section
    str_ += get_determinant_section(protein, conformation, parameters)
    # writing pKa summary section
    str_ += get_summary_section(protein, conformation, parameters)
    str_ += "{0:s}\n".format(get_the_line())
    # printing Folding Profile
    str_ += get_folding_profile_section(
        protein, conformation=conformation, reference=reference,
        window=protein.options.window)
    # printing Protein Charge Profile
    str_ += get_charge_profile_section(protein, conformation=conformation)
    # now, writing the pka text to file
    Path(filename).write_text(str_, encoding="utf-8")


def print_result(protein: "MolecularContainer", conformation: str, parameters: Parameters):
    """Prints all resulting output from determinants and down.

    Args:
        protein:  protein object
        conformation:  specific conformation
        parameters:  parameters
    """
    print_pka_section(protein, conformation, parameters)


def print_pka_section(protein: "MolecularContainer", conformation: str, parameters: Parameters):
    """Prints out pKa section of results.

    Args:
        protein:  protein object
        conformation:  specific conformation
        parameters:  parameters
    """
    # geting the determinants section
    str_ = get_determinant_section(protein, conformation, parameters)
    _LOGGER.info("pKa determinants:\n%s", str_)
    str_ = get_summary_section(protein, conformation, parameters)
    _LOGGER.info("pKa summary:\n%s", str_)


def get_determinant_section(protein: "MolecularContainer", conformation: str, parameters: Parameters):
    """Returns string with determinant section of results.

    Args:
        protein:  protein object
        conformation:  specific conformation
        parameters:  parameters
    Returns:
        string
    """
    # getting the same order as in propka2.0
    str_ = "{0:s}\n".format(get_determinants_header())
    # printing determinants
    for chain in protein.conformations[conformation].chains:
        for residue_type in parameters.write_out_order:
            groups = [
                g for g in protein.conformations[conformation].groups
                if g.atom.chain_id == chain]
            for group in groups:
                if group.residue_type == residue_type:
                    str_ += "{0:s}".format(
                        group.get_determinant_string(
                            parameters.remove_penalised_group))
    # Add a warning in case of coupled residues
    if (protein.conformations[conformation].non_covalently_coupled_groups
            and not protein.options.display_coupled_residues):
        str_ += 'Coupled residues (marked *) were detected.'
        str_ += 'Please rerun PropKa with the --display-coupled-residues \n'
        str_ += 'or -d option for detailed information.\n'
    return str_


def get_summary_section(protein: "MolecularContainer", conformation: str,
                        parameters: Parameters):
    """Returns string with summary section of the results.

    Args:
        protein:  protein object
        conformation:  specific conformation
        parameters:  parameters
    Returns:
        string
    """
    str_ = "{0:s}\n".format(get_summary_header())
    # printing pKa summary
    for residue_type in parameters.write_out_order:
        for group in protein.conformations[conformation].groups:
            if group.residue_type == residue_type:
                str_ += "{0:s}".format(
                    group.get_summary_string(
                        parameters.remove_penalised_group))
    return str_


def get_folding_profile_section(
        protein: "MolecularContainer",
        conformation: str = 'AVR',
        direction: str = "folding",
        reference: str = "neutral",
        window: Tuple[float, float, float] = (0., 14., 1.),
):
    """Returns string with the folding profile section of the results.

    Args:
        protein:  protein object
        conformation:  specific conformation
        direction:  'folding' or other
        reference:  reference state
        window:  pH window [min, max, step]
    Returns:
        string
    """
    str_ = get_the_line()
    str_ += "\n"
    str_ += "Free energy of {0:>9s} (kcal/mol) as a function".format(direction)
    str_ += " of pH (using {0:s} reference)\n".format(reference)
    profile, [ph_opt, dg_opt], [dg_min, dg_max], [ph_min, ph_max] = (
        protein.get_folding_profile(
            conformation=conformation, reference=reference,
            grid=protein.options.grid))

    if profile is None:
        str_ += "Could not determine folding profile\n"
    else:
        delta = round(Decimal(window[2]),2)
        for (ph, dg) in profile:
            ph = round(Decimal(ph), 3)
            if ph >= window[0] and ph <= window[1]:
                if ph % delta < 0.05 or ph % delta > 0.95:
                    str_ += "{0:>6.2f}{1:>10.2f}\n".format(ph, dg)
        str_ += "\n"
    if ph_opt is None or dg_opt is None:
        str_ += "Could not determine pH optimum\n"
    else:
        str_ += "The pH of optimum stability is {0:>4.1f}".format(ph_opt)
        str_ += (
            " for which the free energy is {0:>6.1f} kcal/mol at "
            "298K\n".format(dg_opt)
        )
    if dg_min is None or dg_max is None:
        str_ += "Could not determine pH values where the free energy"
        str_ += " is within 80 % of minimum\n"
    else:
        str_ += "The free energy is within 80 % of maximum"
        str_ += " at pH {0:>4.1f} to {1:>4.1f}\n".format(dg_min, dg_max)
    if ph_min is None or ph_max is None:
        str_ += "Could not determine the pH-range where the free"
        str_ += " energy is negative\n\n"
    else:
        str_ += "The free energy is negative in the range"
        str_ += " {0:>4.1f} - {1:>4.1f}\n\n".format(ph_min, ph_max)
    return str_


def get_charge_profile_section(protein: "MolecularContainer",
                               conformation: str = 'AVR'):
    """Returns string with the charge profile section of the results.

    Args:
        protein:  protein object
        conformation:  specific conformation
        _:  options object
    Returns:
        string
    """
    str_ = "Protein charge of folded and unfolded state as a function of pH\n"
    profile = protein.get_charge_profile(conformation=conformation,
                                         grid=protein.options.grid)
    if profile is None:
        str_ += "Could not determine charge profile\n"
    else:
        str_ += '    pH  unfolded  folded\n'
        for (ph, q_mod, q_pro) in profile:
            str_ += "{ph:6.2f}{qm:10.2f}{qp:8.2f}\n".format(
                ph=ph, qm=q_mod, qp=q_pro)
    pi_pro, pi_mod = protein.get_pi(conformation=conformation)
    if pi_pro is None or pi_mod is None:
        str_ += "Could not determine the pI\n\n"
    else:
        str_ += (
            f"The pI is {pi_pro:>5.2f} (folded) and {pi_mod:>5.2f} "
            f"(unfolded)\n"
        )
    return str_


def get_propka_header():
    """Create the header.

    Returns:
        string
    """
    today = date.today()
    str_ = f"propka{__version__:<53} {today!s:>43}\n"
    str_ += """
-------------------------------------------------------------------------------
--                                                                           --
--  PROPKA: A PROTEIN PKA PREDICTOR                                          --
--                                                                           --
--  VERSION 1.0,  04/25/2004,  IOWA CITY                                     --
--  BY HUI LI                                                                --
--                                                                           --
--  VERSION 2.0,  11/05/2007, IOWA CITY/COPENHAGEN                           --
--  BY DELPHINE C. BAS AND DAVID M. ROGERS                                   --
--                                                                           --
--  VERSION 3.0,  01/06/2011, COPENHAGEN                                     --
--  BY MATS H.M. OLSSON AND CHRESTEN R. SONDERGARD                           --
--                                                                           --
--  VERSION 3.1,  07/01/2011, COPENHAGEN                                     --
--  BY CHRESTEN R. SONDERGARD AND MATS H.M. OLSSON                           --
--                                                                           --
--  VERSION 3.2,  06/17/2020, PLANET EARTH                                   --
--  SEE ABOVE FOR AUTHORS                                                    --
--                                                                           --
-------------------------------------------------------------------------------
"""
    return str_


def get_references_header():
    """Create the 'references' part of output file.

    Returns:
        string
    """
    str_ = """
-------------------------------------------------------------------------------
References:

Very Fast Empirical Prediction and Rationalization of Protein pKa Values.
Hui Li, Andrew D. Robertson and Jan H. Jensen. PROTEINS: Structure, Function,
and Bioinformatics. 61:704-721 (2005)

Very Fast Prediction and Rationalization of pKa Values for Protein-Ligand
Complexes.  Delphine C. Bas, David M. Rogers and Jan H. Jensen.  PROTEINS:
Structure, Function, and Bioinformatics 73:765-783 (2008)

PROPKA3: Consistent Treatment of Internal and Surface Residues in Empirical
pKa predictions.  Mats H.M. Olsson, Chresten R. Sondergard, Michal Rostkowski,
and Jan H. Jensen.  Journal of Chemical Theory and Computation, 7(2):525-537
(2011)

Improved Treatment of Ligands and Coupling Effects in Empirical Calculation
and Rationalization of pKa Values.  Chresten R. Sondergaard, Mats H.M. Olsson,
Michal Rostkowski, and Jan H. Jensen.  Journal of Chemical Theory and
Computation, (2011)
-------------------------------------------------------------------------------
"""
    return str_


def get_warning_header():
    """Create the 'warning' part of the output file.

    TODO - this function is essentially a no-op.

    Returns:
        string
    """
    str_ = ""
    return str_


def get_determinants_header():
    """Create the Determinant header.

    Returns:
        string
    """
    str_ = """
---------  -----   ------   ---------------------    --------------    --------------    --------------
                            DESOLVATION  EFFECTS       SIDECHAIN          BACKBONE        COULOMBIC
 RESIDUE    pKa    BURIED     REGULAR      RE        HYDROGEN BOND     HYDROGEN BOND      INTERACTION
---------  -----   ------   ---------   ---------    --------------    --------------    --------------
"""
    return str_


def get_summary_header():
    """Create the summary header.

    Returns:
        string
    """
    str_ = get_the_line()
    str_ += "\n"
    str_ += "SUMMARY OF THIS PREDICTION\n"
    str_ += "       Group      pKa  model-pKa   ligand atom-type"
    return str_


def get_the_line():
    """Draw the line-Johnny Cash would have been proud-or actually Aerosmith!

    NOTE - Johnny Cash walked the line.

    Returns:
        string
    """
    str_ = ""
    str_ += ("-" * 104)
    return str_


def make_interaction_map(name, list_, interaction):
    """Print out an interaction map named 'name' of the groups in 'list'
    based on the function 'interaction'

    Args:
        list_:  list of groups
        interaction:  some sort of function
    Returns:
        string
    """
    # return an empty string, if the list is empty
    if len(list_) == 0:
        return ''
    # for long list, use condensed formatting
    if len(list_) > 10:
        res = 'Condensed form:\n'
        for i, group1 in enumerate(list_):
            for group2 in list_[i:]:
                if interaction(group1, group2):
                    res += 'Coupling: {0:>9s} - {1:>9s}\n'.format(
                        group1.label, group2.label)
        return res
    # Name and map header
    res = '{0:s}\n{1:>12s}'.format(name, '')
    for group in list_:
        res += '{0:>9s} | '.format(group.label)
    # do the map
    for group1 in list_:
        res += '\n{0:<12s}'.format(group1.label)
        for group2 in list_:
            tag = ''
            if interaction(group1, group2):
                tag = '    X     '
            res += '{0:>10s}| '.format(tag)
    return res


def write_pdb_for_atoms(atoms: List["Atom"], filename: _PathArg, make_conect_section=False):
    """Write out PDB file for atoms.

    Args:
        atoms:  list of atoms
        filename:  name of file
        make_conect_section:  generate a CONECT PDB section
    """
    out = open_file_for_writing(filename)
    for atom in atoms:
        out.write(atom.make_pdb_line())
    if make_conect_section:
        for atom in atoms:
            out.write(atom.make_conect_line())
    out.close()


def get_bond_order(atom1, atom2):
    """Get the order of a bond between two atoms.

    Args:
        atom1:  first atom in bond
        atom2:  second atom in bond
    Returns:
        string with bond type
    """
    type_ = '1'
    pi_electrons1 = atom1.num_pi_elec_2_3_bonds
    pi_electrons2 = atom2.num_pi_elec_2_3_bonds
    if '.ar' in atom1.sybyl_type:
        pi_electrons1 -= 1
    if '.ar' in atom2.sybyl_type:
        pi_electrons2 -= 1
    if pi_electrons1 > 0 and pi_electrons2 > 0:
        type_ = '{0:d}'.format(min(pi_electrons1, pi_electrons2)+1)
    if '.ar' in atom1.sybyl_type and '.ar' in atom2.sybyl_type:
        type_ = 'ar'
    return type_


def write_mol2_for_atoms(atoms, filename):
    """Write out MOL2 file for atoms.

    Args:
        atoms:  list of atoms
        filename:  name of file
    """
    # TODO - header needs to be converted to format string
    header = '@<TRIPOS>MOLECULE\n\n{natom:d} {id:d}\nSMALL\nUSER_CHARGES\n'
    atoms_section = '@<TRIPOS>ATOM\n'
    for i, atom in enumerate(atoms):
        atoms_section += atom.make_mol2_line(i+1)
    bonds_section = '@<TRIPOS>BOND\n'
    id_ = 1
    for i, atom1 in enumerate(atoms):
        for j, atom2 in enumerate(atoms, i+1):
            if atom1 in atom2.bonded_atoms:
                type_ = get_bond_order(atom1, atom2)
                bonds_section += '{0:>7d} {1:>7d} {2:>7d} {3:>7s}\n'.format(
                    id_, i+1, j+1, type_)
                id_ += 1
    substructure_section = '@<TRIPOS>SUBSTRUCTURE\n\n'
    if len(atoms) > 0:
        substructure_section = (
            '@<TRIPOS>SUBSTRUCTURE\n{0:<7d} {1:>10s} {2:>7d}\n'.format(
                atoms[0].res_num, atoms[0].res_name, atoms[0].numb))
    out = open_file_for_writing(filename)
    out.write(header.format(natom=len(atoms), id=id_-1))
    out.write(atoms_section)
    out.write(bonds_section)
    out.write(substructure_section)
    out.close()



"""
Set-up of a PROPKA calculation
==============================

Implements many of the main functions used to call PROPKA.
"""

import logging
import argparse
from pathlib import Path
from typing import Dict, Iterable, Iterator, List, TYPE_CHECKING, NoReturn, Optional, Tuple, TypeVar

if TYPE_CHECKING:
    from propka.atom import Atom
    from propka.conformation_container import ConformationContainer


T = TypeVar("T")
Number = TypeVar("Number", int, float)

_T_RESIDUE_TUPLE = Tuple[str, int, str]

_LOGGER = logging.getLogger(__name__)


EXPECTED_ATOM_NUMBERS = {'ALA': 5, 'ARG': 11, 'ASN': 8, 'ASP': 8, 'CYS': 6,
                         'GLY': 4, 'GLN': 9, 'GLU': 9, 'HIS': 10, 'ILE': 8,
                         'LEU': 8, 'LYS': 9, 'MET': 8, 'PHE': 11, 'PRO': 7,
                         'SER': 6, 'THR': 7, 'TRP': 14, 'TYR': 12, 'VAL': 7}


class Options:
    # Note: All the "NoReturn" members appear to be unused
    chains: Optional[List[str]]
    display_coupled_residues: bool = False
    filenames: List[str]  # List[Path]?
    grid: Tuple[float, float, float] = (0.0, 14.0, 0.1)
    input_pdb: str  # Path?
    keep_protons: bool = False
    log_level: str = 'INFO'
    pH: NoReturn  # float = 7.0
    parameters: Path
    protonate_all: bool = False
    reference: NoReturn  # str = 'neutral'
    reuse_ligand_mol2_file: bool = False  # only used by unused function
    titrate_only: Optional[List[_T_RESIDUE_TUPLE]]
    window: Tuple[float, float, float] = (0.0, 14.0, 1.0)


def protein_precheck(conformations: Dict[str, "ConformationContainer"],
                     names: Iterable[str]):
    """Check protein for correct number of atoms, etc.

    Args:
        names:  conformation names to check
    """
    for name in names:
        atoms = conformations[name].atoms
        # Group the atoms by their residue:
        atoms_by_residue: Dict[str, List[Atom]] = {}
        for atom in atoms:
            if atom.element != 'H':
                res_id = resid_from_atom(atom)
                try:
                    atoms_by_residue[res_id].append(atom)
                except KeyError:
                    atoms_by_residue[res_id] = [atom]
        for res_id, res_atoms in atoms_by_residue.items():
            res_name = res_atoms[0].res_name
            residue_label = '{0:>3s}{1:>5s}'.format(res_name, res_id)
            # ignore ligand residues
            if res_name not in EXPECTED_ATOM_NUMBERS:
                continue
            # check for c-terminal
            if 'C-' in [a.terminal for a in res_atoms]:
                if len(res_atoms) != EXPECTED_ATOM_NUMBERS[res_name]+1:
                    str_ = ("Unexpected number ({num:d}) of atoms in residue "
                            "{res:s} in conformation {conf:s}".format(
                                num=len(res_atoms), res=residue_label,
                                conf=name))
                    _LOGGER.warning(str_)
                continue
            # check number of atoms in residue
            if len(res_atoms) != EXPECTED_ATOM_NUMBERS[res_name]:
                str_ = ("Unexpected number ({num:d}) of atoms in residue "
                        "{res:s} in conformation {conf:s}".format(
                            num=len(res_atoms), res=residue_label,
                            conf=name))
                _LOGGER.warning(str_)


def resid_from_atom(atom):
    """Return string with atom residue information.

    Args:
        atom:  atom to generate string for
    Returns
        string
    """
    return '{0:>4d} {1:s} {2:s}'.format(
        atom.res_num, atom.chain_id, atom.icode)


def split_atoms_into_molecules(atoms: List["Atom"]):
    """Maps atoms into molecules.

    Args:
        atoms:  list of atoms
    Returns:
        list of molecules
    """
    molecules: List[List["Atom"]] = []
    while len(atoms) > 0:
        initial_atom = atoms.pop()
        molecules.append(make_molecule(initial_atom, atoms))
    return molecules


def make_molecule(atom: "Atom", atoms: List["Atom"]):
    """Make a molecule from atoms.

    Args:
        atom:  one of the atoms
        atoms:  a list of the remaining atoms
    Return:
        list of atoms
    """
    bonded_atoms = [a for a in atoms if atom in a.bonded_atoms]
    res_atoms = [atom]
    for bond_atom in bonded_atoms:
        if bond_atom in atoms:
            atoms.remove(bond_atom)
            res_atoms.extend(make_molecule(bond_atom, atoms))
    return res_atoms


def make_grid(min_: Number, max_: Number, step: Number) -> Iterator[Number]:
    """Make a grid across the specified tange.

    Like range() for integers or numpy.arange() for floats, except that `max_`
    is not excluded from the range.

    Args:
        min_:  minimum value of grid
        max_:  maximum value of grid
        step:  grid step size
    """
    x = min_
    while x <= max_:
        yield x
        x += step


def generate_combinations(interactions: Iterable[T]) -> List[List[T]]:
    """Generate combinations of interactions.

    Args:
        interactions:  list of interactions
    Returns:
        list of combinations
    """
    res: List[List[T]] = [[]]
    for interaction in interactions:
        res = make_combination(res, interaction)
    res.remove([])
    return res


def make_combination(combis: List[List[T]], interaction: T) -> List[List[T]]:
    """Make a specific set of combinations.

    Args:
        combis:  list of combinations
        interaction:  interaction to add to combinations
    Returns:
        list of combinations
    """
    res = []
    for combi in combis:
        res.append(combi+[interaction])
        res.append(combi)
    return res


def parse_res_string(res_str: str) -> _T_RESIDUE_TUPLE:
    """Parse a residue string.

    Args:
        res_string:  residue string in format "chain:resnum[inscode]"
    Returns:
        a tuple of (chain, resnum, inscode).
    Raises:
        ValueError if the input string is invalid.
    """
    try:
        chain, resnum_str = res_str.split(":")
    except ValueError:
        raise ValueError("Invalid residue string (must contain 2 "
                         "colon-separated values)")
    try:
        resnum = int(resnum_str)
    except ValueError:
        try:
            resnum = int(resnum_str[:-1])
        except ValueError:
            raise ValueError("Invalid residue number (not an int)")
        else:
            inscode = resnum_str[-1]
    else:
        inscode = " "
    return chain, resnum, inscode


def parse_res_list(titrate_only: str):
    res_list: List[_T_RESIDUE_TUPLE] = []
    for res_str in titrate_only.split(','):
        try:
            res_list.append(parse_res_string(res_str))
        except ValueError as ex:
            raise argparse.ArgumentTypeError(f'{ex}: "{res_str:s}"')
    return res_list


def build_parser(parser=None):
    """Build an argument parser for PROPKA.

    Args:
        parser:  existing parser. If this is not None, then the PROPKA parser
                 will be created as a subparser to this existing parser.
                 Otherwise, a new parser will be created.
    Returns:
        ArgumentParser object.


    .. versionchanged:: 3.4.0
       Argument `--generate-propka-input` has been removed as writing PROPKA
       input files is no longer supported.
    """
    import propka

    if parser is not None:
        group = parser.add_argument_group(title="PROPKA invocation options")
    else:
        parser = argparse.ArgumentParser(
            description=("PROPKA predicts the pKa values of ionizable "
                         "groups in proteins and protein-ligand "
                         "complexes based in the 3D structure"),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        # This is duck-typing at its finest
        group = parser
        group.add_argument("input_pdb", help="read data from <filename>")
    group.add_argument(
        "-f", "--file", action="append", dest="filenames", default=[],
        help=(
            "read data from <filename>, i.e. <filename> is added to arguments"
        )
    )
    group.add_argument(
        "-r", "--reference", dest="reference", default="neutral",
        help=("setting which reference to use for stability calculations "
              "[neutral/low-pH]"))
    group.add_argument(
        "-c", "--chain", action="append", dest="chains",
        help=('creating the protein with only a specified chain. Specify '
              '" " for chains without ID [all]'))
    group.add_argument(
        "-i", "--titrate_only", dest="titrate_only",
        type=parse_res_list,
        help=('Treat only the specified residues as titratable. Value should '
              'be a comma-separated list of "chain:resnum" values; for '
              'example: -i "A:10,A:11"'))
    group.add_argument(
        "--version", action="version", version=f"%(prog)s {propka.__version__}")
    group.add_argument(
        "-p", "--parameters", dest="parameters",
        type=Path, default=Path(__file__).parent / "propka.cfg",
        help="set the parameter file")
    try:
        group.add_argument(
            "--log-level",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="logging level verbosity", default="INFO")
    except argparse.ArgumentError:
        # It is possible that --log-level has already been set by APBS
        pass
    group.add_argument(
        "-o", "--pH", dest="pH", type=float, default=7.0,
        help="setting pH-value used in e.g. stability calculations [7.0]")
    group.add_argument(
        "-w", "--window", dest="window", nargs=3, type=float,
        default=(0.0, 14.0, 1.0),
        help=("setting the pH-window to show e.g. stability profiles "
              "[0.0, 14.0, 1.0]"))
    group.add_argument(
        "-g", "--grid", dest="grid", nargs=3, type=float,
        default=(0.0, 14.0, 0.1),
        help=("setting the pH-grid to calculate e.g. stability "
              "related properties [0.0, 14.0, 0.1]"))
    group.add_argument(
        "-d", "--display-coupled-residues", dest="display_coupled_residues",
        action="store_true",
        help=("Displays alternative pKa values due "
              "to coupling of titratable groups"))
    group.add_argument(
        "-l", "--reuse-ligand-mol2-files", dest="reuse_ligand_mol2_file",
        action="store_true", default=False,
        help=("Reuses the ligand mol2 files allowing the user to alter "
              "ligand bond orders"))
    group.add_argument(
        "-k", "--keep-protons", dest="keep_protons", action="store_true",
        help="Keep protons in input file", default=False)
    group.add_argument(
        "-q", "--quiet", action="store_const", const="WARNING",
        dest="log_level", help="suppress non-warning messages")
    group.add_argument(
        "--protonate-all", dest="protonate_all", action="store_true",
        help="Protonate all atoms (will not influence pKa calculation)",
        default=False)
    return parser


def loadOptions(args=None) -> Options:
    """
    Load the arguments parser with options. Note that verbosity is set as soon
    as this function is invoked.

    Arguments:
        args:  list of arguments
    Returns:
        argparse namespace
    """
    # loading the parser
    parser = build_parser()
    # parsing and returning options and arguments
    options = parser.parse_args(args, namespace=Options())

    # adding specified filenames to arguments
    options.filenames.append(options.input_pdb)
    # Set the no-print variable
    level = getattr(logging, options.log_level)
    _LOGGER.setLevel(level)
    # done!
    return options


def make_tidy_atom_label(name, element):
    """Returns a 'tidier' atom label for printing to the new PDB file.

    Args:
        name:  atom name
        element:  atom element
    Returns:
        string
    """
    if len(name) > 4:  # if longer than 4, just truncate the name
        label = name[0:4]
    elif len(name) == 4:  # if length is 4, otherwise use the name as it is
        label = name
    else:  # if less than 4 characters long, insert white space as needed
        if len(element) == 1:
            label = ' {0:<3s}'.format(name)
        else:  # The element should occupy the two first chars
            label = '{0:<4s}'.format(name)
    return label



"""
Hybrid36 PDB-like file format
=============================

`hybrid36`_ is an alternative PDB format that can encode larger atom
numbers. This module provides the :func:`decode` functon to parse the
atom numbers in hybrid36 "PDB" files.

.. _hybrid36: http://cci.lbl.gov/hybrid_36/

"""
import string


_HYBRID36_UPPER_CHARS = set(string.ascii_uppercase)
_HYBRID36_LOWER_CHARS = set(string.ascii_lowercase)
_HYBRID36_DIGITS = set(string.digits)
_HYBRID36_UPPER_SET = _HYBRID36_UPPER_CHARS | _HYBRID36_DIGITS
_HYBRID36_LOWER_SET = _HYBRID36_LOWER_CHARS | _HYBRID36_DIGITS


def decode(input_string):
    """Convert an input string of a number in hybrid-36 format to an integer.

    Args:
        input_string:  input string
    Returns:
        integer
    """
    value_error_message = "invalid literal for hybrid-36 conversion: '{0:s}'"

    original_input_string = input_string
    input_string = input_string.strip()

    # Manually handle negative sign.
    if input_string.startswith("-"):
        sign = -1
        input_string = input_string[1:]
    else:
        sign = 1

    if len(input_string) == 0:
        raise ValueError(value_error_message.format(input_string))

    # See http://cci.lbl.gov/hybrid_36/ for documentation on the format.

    num_chars = len(input_string)
    first_char = input_string[0]

    if first_char in _HYBRID36_DIGITS:
        return sign * int(input_string)
    elif first_char in _HYBRID36_UPPER_CHARS:
        reference = - (10 * 36 ** (num_chars - 1) - 10 ** num_chars)
        _hybrid36_set = _HYBRID36_UPPER_SET
    elif first_char in _HYBRID36_LOWER_CHARS:
        reference = (16 * 36 ** (num_chars - 1) + 10 ** num_chars)
        _hybrid36_set = _HYBRID36_LOWER_SET
    else:
        raise ValueError(value_error_message.format(original_input_string))

    # Check the validity of the input string: ASCII characters should be
    # either all uppercase or all lowercase.
    for char in input_string[1:]:
        if char not in _hybrid36_set:
            raise ValueError(value_error_message.format(original_input_string))

    # Convert with the int function.
    return sign * (int(input_string, 36) + reference)



"""
Script functionality
====================

The :mod:`run` module provides a high-level interface to PROPKA 3.

The :program:`propka3` script consists of the :func:`main`
function. If similar functionality is desired from a Python script
(without having to call the :program:`propka` script itself) then the
:func:`single` function can be used instead.

"""
import logging
import sys
from typing import IO, Iterable, Optional
from propka.lib import loadOptions
from propka.input import read_parameter_file, read_molecule_file
from propka.parameters import Parameters
from propka.molecular_container import MolecularContainer
from propka.output import _PathArg


_LOGGER = logging.getLogger(__name__)


def main(optargs=None):
    """Read in structure files, calculate pKa values, and print pKa files.


    .. versionchanged:: 3.4.0
       Removed ability to write out PROPKA input files.
    """
    # loading options, flags and arguments
    logger = logging.getLogger("")
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(logging.Formatter("%(message)s"))
    logger.addHandler(stdout_handler)
    optargs = optargs if optargs is not None else []
    options = loadOptions(*optargs)
    pdbfiles = options.filenames
    parameters = read_parameter_file(options.parameters, Parameters())
    for pdbfile in pdbfiles:
        my_molecule = MolecularContainer(parameters, options)
        my_molecule = read_molecule_file(pdbfile, my_molecule)
        my_molecule.calculate_pka()
        my_molecule.write_pka()


def single(filename: _PathArg,
           optargs: Iterable[str] = (),
           stream: Optional[IO[str]] = None,
           write_pka: bool = True):
    """Run a single PROPKA calculation using ``filename`` as input.

    Args:
        filename (str): name of input file. If filestream is not passed via
            ``stream``, should be a path to the file to be read.
        optargs (tuple): Optional, commandline options for propka. Extra files
            passed via ``optargs`` will be ignored, see Notes.
        stream : optional filestream handle. If ``None``, then ``filename``
            will be used as path to input file for reading.
        write_pka (bool): Controls if the pKa file should be writen to disk.

    Returns:
        :class:`~propka.molecular_container.MolecularContainer` object.

    Examples:
        Given an input file "protein.pdb", run the equivalent of ``propka3
        --mutation=N25R/N181D -v --pH=7.2 protein.pdb`` as::

            propka.run.single("protein.pdb",
                optargs=["--mutation=N25R/N181D", "-v", "--pH=7.2"])

        By default, a pKa file will be written. However in some cases one may
        wish to not output this file and just have access to the
        :class:`~propka.molecular_container.MolecularContainer` object. If so,
        then pass ``False`` to ``write_pka``::

            mol = propka.run.single("protein.pdb", write_pka=False)

        In some cases, one may also want to pass a file-like (e.g.
        :class:`io.StringIO`) object instead of a file path as a string. In
        these cases the file-like object should be passed to the ``stream``
        argument and a string indicating the file type in the ``filename``
        argument; this string only has to look like a valid file name, it does
        not need to exist because the data are actually read from ``stream``.
        This approach is necessary because file-like objects do not usually
        have names, and propka uses the ``filename`` argument  to determine the
        input file type, and assigns the file name for the
        :class:`~propka.molecular_container.MolecularContainer` object::

            mol = propka.run.single('input.pdb', stream=string_io_file)

        In this case, a PDB file-like object was passed as `string_io_file`.
        The resultant pKa file will be written out as `input.pka`.

    Notes:
        * Only a single input structure file will be processed, defined by
          ``filename`` (and ``stream`` if passing a file-like object). Any
          additional files passed via the `-f` or `--file` flag to optargs will
          be ignored.


    .. seealso::

        :func:`propka.input.read_molecule_file`


    .. versionchanged:: 3.4.0
       Removed ability to write out PROPKA input files.
    """
    filename = str(filename)
    # Deal with input optarg options
    optargs = tuple(optargs)
    optargs += (filename,)
    options = loadOptions(optargs)

    parameters = read_parameter_file(options.parameters, Parameters())

    # Only filename present should be the one passed via the arguments
    # Anything else will probably have been passed using optargs' `-f` flag.
    ignored_list = [i for i in options.filenames if i != filename]
    if ignored_list:
        _LOGGER.warning(f"Ignoring extra filenames passed: {ignored_list}")
    options.filenames = [filename]

    my_molecule = MolecularContainer(parameters, options)
    my_molecule = read_molecule_file(filename, my_molecule, stream=stream)
    my_molecule.calculate_pka()

    # write outputs
    if write_pka:
        my_molecule.write_pka()

    return my_molecule
